--------------------------------------------------------------------------
--
-- IDECPLD 
--
-- IDE Interface for DBox2
--
-- (C) 2006 Uli Tessel (aka DboxBaer)
-- (C) 2008 kpt.ahab & Bastian Rehberg (aka Stingray) fixed for Sagem
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; (version 2 of the License)
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
--
--------------------------------------------------------------------------
--
-- CPUStates:   
--
-- handles CPU Accesses (without looking at the data)
-- the address is used to find what the driver wants to do, see
-- table below
-- 
--------------------------------------------------------------------------
library IEEE;

use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity CPUStates is
  Port (
    MemCmd:   in  STD_LOGIC_VECTOR (3 downto 0);
	 Addr:     in  STD_LOGIC_VECTOR (9 downto 2);
    dataOE:   out STD_LOGIC;
	 dataMux:  out STD_LOGIC_VECTOR (1 downto 0);	
	 dataLoad: out STD_LOGIC;
	 readFromFifo: out STD_LOGIC;
    writeHighToFifo: out STD_LOGIC;
    writeLowToFifo: out STD_LOGIC;
    writeToCtrl: out STD_LOGIC;
	 writeToTiming: out STD_LOGIC;	 
	 Reset:    in  STD_LOGIC;
    clk :     in  STD_LOGIC );
    end CPUStates;

architecture Behavioral of CPUStates is

  type cpu_state_type is (
    cs_idle, 
    cs_ras1, cs_ras2,  cs_rasw, 
    cs_rd1, cs_rd2,
    cs_wr1); 
	 
  signal cpu_state, cpu_next_state : cpu_state_type; 

begin

  process (clk, Reset)
  begin
    if (Reset = '0') then
        cpu_state <= cs_idle;
    elsif (clk'event and clk = '1') then
        cpu_state <= cpu_next_state;
    end if;        
  end process;



  
process (clk, cpu_state, MemCmd)
   begin
     case (cpu_state) is
       when cs_idle => if (MemCmd = 5) then
                         cpu_next_state <= cs_ras1;
                       else
                         cpu_next_state <= cs_idle;
                       end if;
       when cs_ras1 => cpu_next_state <= cs_ras2;
   
        when cs_ras2 => if (MemCmd = 7) then
                         cpu_next_state <= cs_rasw;
         	       elsif (MemCmd = 2) then  
        		 cpu_next_state <= cs_wr1;
                       elsif (MemCmd = 6) then
                         cpu_next_state <= cs_rd1;
                       else
                         cpu_next_state <= cs_idle;
                       end if;
 
       when cs_rasw => if (MemCmd = 2) then
                         cpu_next_state <= cs_wr1;
                       else
                         cpu_next_state <= cs_idle;
       end if;
 
       when cs_wr1 =>  cpu_next_state <= cs_idle;
 
       when cs_rd1 =>  cpu_next_state <= cs_rd2;
       when cs_rd2 =>  cpu_next_state <= cs_idle;
       when others =>  cpu_next_state <= cs_idle;
     end case;
   end process;

  
    -- output driver are enabled when cpu reads
  
  -- alternative A: only in one cycle:
  --dataOE <= '1' when (cpu_state = cs_rd2) else '0';
  
  -- alternative B: during two cycles. 
  -- This relaxes the bus-timing a lot!
  dataOE <= '1' when ((cpu_state=cs_rd1) or
                      (cpu_state=cs_rd2)) else '0';
  
 
  -- new data is loaded after ras2 -> this is either 
  -- when the cpu writes or when it reads...
  dataLoad <= '1' when (cpu_state=cs_ras2) and Addr(9)='1' else '0';
  
  -- ... so the address is used to decide what is written to the 
  -- data register
  dataMux <= Addr(8) & Addr(7);
  
  -- Address usage:
  -- A9 A8 A7 A6 A5 A4 A3
  --  0  x  x  x  x  x  x don't change data (even on write). Can be used to read data-reg
  --  1  0  0  x  x  x  x cpu write to data ...
  --  x  x  x  1  x  x  x   during write-access: store high word to fifo
  --  x  x  x  x  1  x  x   during write-access: store low word to fifo
  --  x  x  x  0  0  1  x   during write-access: store data in ctrl
  --  x  x  x  0  0  x  1   during write-access: store data in timing
  --  1  0  1  x  x  x  x read data from fifo to data
  --  1  1  0  x  x  x  x read data from ctrl
 
  readFromFifo <= '1' when (cpu_state=cs_rd1) and (Addr(9 downto 7)="101") else '0';
  
  process (clk,reset)
  begin
    if (Reset = '0') then
	    writeHighToFifo <= '0';
	    writeLowToFifo  <= '0';
    elsif (clk'event and clk = '1') then
	 
	   if (cpu_state=cs_wr1) then
	    writeHighToFifo <= Addr(6);
	    writeLowToFifo  <= Addr(5);
      else
	    writeHighToFifo <= '0'; -- todo: can be cleared by fifo?!
	    writeLowToFifo  <= '0';
      end if;		
		
    end if;        
  end process;
  
  writeToCtrl     <= '1' when (cpu_state=cs_wr1) and (Addr(6 downto 4)="001") and (Addr(2)='0') else '0';
  writeToTiming   <= '1' when (cpu_state=cs_wr1) and (Addr(6 downto 5)="00") and (Addr(3 downto 2)="10") else '0';

   
end Behavioral;
