--------------------------------------------------------------------------
--
-- IDECPLD 
--
-- IDE Interface for DBox2
--
-- (C) 2006 Uli Tessel (aka DboxBaer)
-- (C) 2008 kpt.ahab & Bastian Rehberg (aka Stingray) fixed for Sagem
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; (version 2 of the License)
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
--
--------------------------------------------------------------------------
--
-- fifo:   
-- cpu can read and write 2 words (=2*16Bit) at once
-- ide reads and writes single words
-- 
-- 4 16 Bit D-Flipflops (f0-f3) are used with 4 additional 
-- "data-available" (da) D-Flipflops that represent a
-- word is available in "their" FlipFlops.
-- 
-- when there is data in the fifo it moves automatically to the front.
--
-- Byte Order:
-- (Independent from PowerPC definition: in this chip 
-- D31 is 2^31, D7 is 2^7=128, D0 is 2^0=1)
-- 
-- The data in a sector on a disk are organized in bytes. Obviously
-- the first byte should be stored as the first byte in memory, 
-- and so on. 
-- So this defines the byte order and how the CPU Bus is connected
-- to this fifo:
--
-- The first byte read from the data register of a disk is found in 
-- IDEData(7..0) the second byte is in IDEData(15..8)
-- 
-- The PPC-CPU reads two words from the fifo at once and places 
-- D(31..24) as first byte to memory, D(23..16) as second, 
-- D(15..8) as third and D(7..0) is the fourth byte
--
--------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity Fifo is 
  Port (
	 cpuIn: in STD_LOGIC_VECTOR(31 downto 0);
	 ideIn: in STD_LOGIC_VECTOR(15 downto 0);

	 cpuRd, cpuWrHi, cpuWrLo: in STD_LOGIC;
	 ideRd, ideWr: in STD_LOGIC;

    hasData: out STD_LOGIC;	 
	 ready: out STD_LOGIC;
	 
	 ideOut: out STD_LOGIC_VECTOR(15 downto 0);
	 cpuOut: out STD_LOGIC_VECTOR(31 downto 0);
	 
	 level: out STD_LOGIC_VECTOR(3 downto 0);

	 reset: in  STD_LOGIC;
    clk : in  STD_LOGIC );
end Fifo;

architecture Behavioral of Fifo is
  
  signal da, nda: STD_LOGIC_VECTOR(3 downto 0);

  signal ce3,ce2,ce1,ce0: STD_LOGIC;
  signal m3,m2: STD_LOGIC;
    
  signal mux3: STD_LOGIC_VECTOR(15 downto 0);
  signal mux2: STD_LOGIC_VECTOR(15 downto 0);

  signal f3: STD_LOGIC_VECTOR(15 downto 0) := x"0000";
  signal f2: STD_LOGIC_VECTOR(15 downto 0) := x"0000";
  signal f1: STD_LOGIC_VECTOR(15 downto 0) := x"5469";
  signal f0: STD_LOGIC_VECTOR(15 downto 0) := x"6C55";

begin

  process (clk, Reset, nda(3),nda(2),nda(1),nda(0))
  begin
    if (Reset = '0') then
      da  <= (others => '0');
    elsif (clk'event and clk = '1') then	 
      da <= nda;
 	 end if;
  end process;
  
  p3: process (da,ce2,cpuWrLo,ideWr)
  begin
    if (da(3)='0') then
	   if    (cpuWrLo= '1') then nda(3)<='1'; 
		elsif (ideWr  = '1') then nda(3)<='1'; 
      else                      nda(3)<='0'; 
      end if;		                      
	 else                                
	   if (ce2='1') then         nda(3)<='0'; 
		else                      nda(3)<='1';  
		end if;
	 end if;
	 ce3 <= cpuWrLo or ideWr;
	 m3  <= cpuWrLo;
  end process;

  p2: process (da,ce1,cpuWrHi)
  begin
    if (da(2)='0') then
	   if    (cpuWrHi= '1') then nda(2)<='1'; 
		elsif (da(3)  = '1') then nda(2)<='1'; 
      else                      nda(2)<='0'; 
      end if;		                       
	 else                                 
	   if    (ce1    = '1') then nda(2)<='0'; 
      else                      nda(2)<='1'; 
      end if;		
	 end if;
	 ce2 <= cpuWrHi or ((not da(2)) and da(3));
	 m2  <= cpuWrHi;
  end process;

  p1: process (da,ce0,cpuRd)
  begin
    if (da(1)='0') then
	   if    (da(2)  = '1') then nda(1)<='1';  
      else                      nda(1)<='0'; 
      end if;		
	 else
	   if    (ce0    = '1') then nda(1)<='0'; 
		elsif (cpuRd  = '1') then nda(1)<='0'; 
      else                      nda(1)<='1';  
      end if;		
	 end if;
	 ce1 <= (not da(1)) and da(2);
  end process;

  p0: process (da,cpuRd,ideRd)
  begin
    if (da(0)='0') then
	   if    (da(1)  = '1') then nda(0)<='1'; 
      else                      nda(0)<='0'; 
      end if;		
	 else
	   if    (ideRd  = '1') then nda(0)<='0'; 
		elsif (cpuRd  = '1') then nda(0)<='0'; 
      else                      nda(0)<='1'; 
      end if;		
	 end if;
	 ce0 <= (not da(0)) and da(1);
  end process;
  
  mux3(15 downto 8) <= cpuIn(7 downto 0) when (m3='1') else ideIn(15 downto 8);
  mux3(7 downto 0)  <= cpuIn(15 downto 8) when (m3='1') else ideIn(7 downto 0) ;
  mux2(15 downto 8) <= cpuIn(23 downto 16)  when (m2='1') else f3(15 downto 8);
  mux2(7 downto 0)  <= cpuIn(31 downto 24)  when (m2='1') else f3(7 downto 0) ;
  
  process (clk,reset,ce3,mux3)
  begin
    if (reset='0') then
	   f3 <= (others => '0');
	 elsif (clk'event and clk='1') then
	   if (ce3='1') then
		  f3 <= mux3;
   	end if;	 
	 end if;
  end process;
      
  process (clk,reset,ce2,mux2)
  begin
    if (reset='0') then
	   f2 <= (others => '0');
	 elsif (clk'event and clk='1') then
	   if (ce2='1') then
		  f2 <= mux2;
   	end if;	 
	 end if;
  end process;

  process (clk,reset,ce1,f2)
  begin
    if (reset='0') then
	   f1 <= (others => '0');
	 elsif (clk'event and clk='1') then
	   if (ce1='1') then
		  f1 <= f2;
   	end if;	 
	 end if;
  end process;
  
  process (clk,reset,ce0,mux3)
  begin
    if (reset='0') then
	   f0 <= (others => '0');
	 elsif (clk'event and clk='1') then
	   if (ce0='1') then
		  f0 <= f1;
   	end if;	 
	 end if;
  end process;
  
  -- there are no resources to create a real level, so the driver get this raw version
  level <= da;
      
  ideOut <= f0;
  
  cpuOut(31 downto 24)<= f0(7 downto 0);
  cpuOut(23 downto 16)<= f0(15 downto 8);
  cpuOut(15 downto 8) <= f1(7 downto 0);
  cpuOut(7 downto 0)  <= f1(15 downto 8);
  
  ready   <= not da(3);
  hasData <=     da(0);

end Behavioral;
