--------------------------------------------------------------------------
--
-- IDECPLD 
--
-- IDE Interface for DBox2
--
-- (C) 2006 Uli Tessel (aka DboxBaer)
-- (C) 2008 kpt.ahab & Bastian Rehberg (aka Stingray) fixed for Sagem
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; (version 2 of the License)
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
--
--------------------------------------------------------------------------
--
-- ide:   
--
-- generates IDE Accesses: The driver configures this part using a
-- ctrl register, then the IDE-Access is executed when the fifo is
-- ready/has data.
-- 
-- Features:
-- This implementation allows to control the full range of Pulse Widths
-- and Cycle Lengths required for different PIO Modes.
--
-- Interrupts are inverted, so the falling edge can be used
-- 
--
-- todo: start with a small delay (when enable is written), so the 
-- access does not start immediately, so the IDEAx and IDECSx can 
-- settle.
-- (but this can also be done by the driver: first write the value 
-- without enable active, then write again to activate it)
--
--------------------------------------------------------------------------
--
-- Usage:
--
-- Ctrl Bits are as follows
--  ctrl(24 downto 20): (cycleDelay in ns)/15-3 
--  ctrl(19 downto 16): (pulseLength in ns)/15-2
--  ctrl(8)           : not IDEDMACK
--  ctrl(7)           : repeat
--  ctrl(6)           : enable
--  ctrl(5)           : writing
--  ctrl(4)           : IDECS1 
--  ctrl(3)           : IDECS0 
--  ctrl(2 downto 0)  : IDEAddr
--
--  0111 0xxx
--  71,72,73,74,75,76,77 -> Single write to IDE Register 1..7
--  
--  0101 0xxx
--  51,52,53,54,55,56,57 -> Single read from IDE Register 1..7
-- 
--  1111 0000
--  F0 -> Write Words to IDE Register 0 (Data)
--
--  1101 0000
--  D0 -> Read Words from IDE Register 0
--
--  0110 1xxx
--  68..6F -> Single write with CS1 low
--
--  0100 1xxx
--  48..4F -> Single read with CS1 low
-- 
--------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity ide is
  Port(
    ctrlIn: in STD_LOGIC_VECTOR(12 downto 0);
    timeIn: in STD_LOGIC_VECTOR(8 downto 0);
    ctrlOut: out STD_LOGIC_VECTOR(12 downto 0);
	 timeOut: out STD_LOGIC_VECTOR(8 downto 0);
	 loadDataToCtrl: in STD_LOGIC;
	 loadDataToTiming: in STD_LOGIC;
	 fifoHasData: in STD_LOGIC;
	 fifoIsReady: in STD_LOGIC;

	 readFromFifo: out STD_LOGIC;
	 writeToFifo: out STD_LOGIC;

	 IDEIntIn: in STD_LOGIC;
	 IDEIntOut: out STD_LOGIC;
    	 	 
    IDEIORDY: in STD_LOGIC;

    IDEAddrOut : out STD_LOGIC_VECTOR (2 downto 0);
    IDECS0Out  : out STD_LOGIC;
    IDECS1Out  : out STD_LOGIC;
    IDEResetOut: out STD_LOGIC;
    IDEDMACKOut: out STD_LOGIC;
	 
	 ideOE: out STD_LOGIC;
	 IDEIOW: out STD_LOGIC;
	 IDEIOR: out STD_LOGIC;
	 

	 ideTSOut: out STD_LOGIC;

	 reset: in STD_LOGIC;
    clk : in STD_LOGIC );
end ide;

architecture Behavioral of ide is
 
  signal cycleDelay: STD_LOGIC_VECTOR(4 downto 0) := "11111";
  signal pulseLength: STD_LOGIC_VECTOR(3 downto 0):="1111";
  
  signal repeat: STD_LOGIC                       := '0';
  signal enable: STD_LOGIC                       := '0';
  signal writing: STD_LOGIC                      := '0';
  signal IDEReset: STD_LOGIC                     := '0';
  signal IDECS1  : STD_LOGIC                     := '1';
  signal IDECS0  : STD_LOGIC                     := '1';
  signal IDEAddr : STD_LOGIC_VECTOR (2 downto 0) := "000";
  signal IDEDMACK: STD_LOGIC                     := '1';
    
  signal delay: STD_LOGIC_VECTOR(4 downto 0)     := "00000";
  signal ldPulse: STD_LOGIC;
  signal ldCycle: STD_LOGIC;
  
  signal state: STD_LOGIC_VECTOR (3 downto 0)    := "0000";
  


begin
  IDEAddrOut  <= IDEAddr;
  IDECS0Out   <= IDECS0;
  IDECS1Out   <= IDECS1;
  IDEResetOut <= IDEReset;
  IDEDMACKOut <= IDEDMACK;  

  ideTSOut    <= not IDEReset; -- tristate the other pins during reset

  process (clk,reset,loadDataToTiming,loadDataToCtrl,ctrlIn,state,repeat)
  begin
    if (Reset = '0') then
	   cycleDelay <= (others => '1');
		pulseLength <= (others => '1');
		
		repeat  <= '0';
		enable  <= '0';
		writing <= '0';
		IDEReset<= '0';
		IDECS1  <= '1';
		IDECS0  <= '1';
		IDEDMACK<= '1';
	   IDEAddr <= "000";
    elsif (clk'event and clk = '1') then

      if (loadDataToTiming='1') then		
		  cycleDelay <= timeIn(8 downto 4);
		  pulseLength<= timeIn(3 downto 0);
		end if;
		
	   if (loadDataToCtrl='1') then
			
		  repeat  <= ctrlIn(7);		  		  

		  IDEDMACK <= not ctrlIn(8);
		  enable  <= ctrlIn(6);
		  writing <= ctrlIn(5);
		  
		  case (ctrlIn(4 downto 3)) is
		  -- use combination of CS0 and CS1 for Reset
		  when "00"   => IDECS0<='1'; IDECS1<='1'; IDEReset<='0';
		  when "01"   => IDECS0<='1'; IDECS1<='0'; IDEReset<='1';
		  when "10"   => IDECS0<='0'; IDECS1<='1'; IDEReset<='1';
		  when others => IDECS0<='1'; IDECS1<='1'; IDEReset<='1';
		  end case;
		  		  
        IDEAddr <= ctrlIn(2 downto 0);
		end if;
		
		if (state(0) = '1') then  
		  enable <= repeat; 
		end if;
			
    end if;        
  end process;
  
  process (clk,reset,delay,ldPulse,ldCycle)
  begin
    if (reset='0') then
	   delay <= (others => '0');
    elsif (clk'event and clk='1') then	 
	   if    (ldPulse='1') then delay <= "0" & pulseLength;
	   elsif (ldCycle='1') then delay <= cycleDelay;
		elsif (delay/=0)    then 
		  delay <= delay-1;		  
	   end if;	 		
	 end if;
  end process;
  
  process (clk, reset,state,delay,enable,writing,fifoHasData,fifoIsReady,IDEIORDY)
  begin
    if (Reset = '0') then  
	   IDEIOW <= '1';
		IDEIOR <= '1';	 
		state <= "0000";
		
    elsif (clk'event and clk = '1') then 	 
	 
	   if (delay=0) and 
		   (state(2)='0') and 
		   (enable='1') and			
			( ((writing = '1') and (fifoHasData='1')) or
			  ((writing = '0') and (fifoIsReady='1')) )
		then
		  state(3) <= '1';
		  state(2) <= '1';
		  
		  IDEIOW <= not writing;
		  IDEIOR <= writing;
		end if;
		
		if (state(3)='1') then
		  state(3) <= '0';
		end if;
		
		if (delay=0) and (state(3)='0') and (state(2)='1') and ((IDEIORDY='1') or ( (IDECS0='1') and (IDECS1='1') )) then
		  state(1) <= '1';
		end if;
		
		if (state(2)='1') and (state(1)='1') and (state(0)='1') then
		  state(1) <= '0';
		  state(2) <= '0';
		end if;
		
		state(0) <= state(1);
		
		if (state(1)='1') then
		  IDEIOW <= '1';
		  IDEIOR <= '1';
		end if;
		
	 end if;
  end process;
  
  timeOut(8 downto 4) <= cycleDelay; 
  timeOut(3 downto 0) <= pulseLength;
    
  ctrlOut(12) <= '0';
  ctrlOut(11) <= '0';
  ctrlOut(10) <= '0';
  ctrlOut(9)  <= '0';
  ctrlOut(8)  <= not IDEDMACK;
  ctrlOut(7)  <= repeat;
  ctrlOut(6)  <= enable;
  ctrlOut(5)  <= writing;  
  ctrlOut(4)  <= IDECS1 and IDEReset;
  ctrlOut(3)  <= IDECS0 and IDEReset;  
  ctrlOut(2 downto 0) <= IDEAddr;
     
  -- the "state" is almost identical to the generated signals
  ideOE  <= state(2) and writing;
  readFromFifo <= state(0) and not state(1) and writing;
  writeToFifo  <= not state(0) and state(1) and not writing;
  ldPulse <= state(3);
  ldCycle <= state(1);  

  -- interrupt is inverted because the 823 wants a falling edge
  IDEIntOut <= not IDEIntIn;

end Behavioral;
