--------------------------------------------------------------------------
--
-- IDECPLD 
--
-- IDE Interface for DBox2
--
-- (C) 2006 Uli Tessel (aka DboxBaer)
-- (C) 2008 kpt.ahab & Bastian Rehberg (aka Stingray) fixed for Sagem
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; (version 2 of the License)
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
--
--------------------------------------------------------------------------
--
-- Main (top level design)
--
-- connects the different parts in the CPLD:
-- 
-- cpuStates: handles the timing of a CPU Acesses
-- dataReg:   the central data register the CPU interacts with
-- fifo:      decouple the IDE part from the CPU
-- ide:       generates IDE accesses
--
--------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity main is
  Port (
  -- CPU Interface
  CPUData : inout STD_LOGIC_VECTOR (31 downto 0);
  CPUAddr : in    STD_LOGIC_VECTOR (9 downto 2);
  MemCmd  : in    STD_LOGIC_VECTOR (3 downto 0); -- CS, WE, RAS, CAS
  CPUInt  : out   STD_LOGIC;

  -- IDE Interface
  IDEData : inout STD_LOGIC_VECTOR (15 downto 0);
  IDEAddr : out   STD_LOGIC_VECTOR (2 downto 0);
  IDECS0  : out   STD_LOGIC;
  IDECS1  : out   STD_LOGIC;
  IDEReset: out   STD_LOGIC;
  IDEDMACK: out   STD_LOGIC;
  IDEIORDY: in    STD_LOGIC;
  IDEIOR  : out   STD_LOGIC;
  IDEIOW  : out   STD_LOGIC;
  IDEIntRq: in    STD_LOGIC;
  IDEDMARq: in    STD_LOGIC;  
  
  
  
  Led_Input: in     STD_LOGIC;
  Led_Output: out    STD_LOGIC; 
  
  
  
  
--reset: in  STD_LOGIC; -- reset is not connected because it is not available
  clk : in  STD_LOGIC );
end main;

architecture Behavioral of main is

  signal Reset: STD_LOGIC; -- reset is "generated" internally
  
  component cpuStates Port (
    MemCmd: in  STD_LOGIC_VECTOR (3 downto 0);
	 Addr:   in  STD_LOGIC_VECTOR (9 downto 2);
	 
    dataOE: out STD_LOGIC;
	 dataMux:  out STD_LOGIC_VECTOR (1 downto 0);	
	 dataLoad: out STD_LOGIC;
	 readFromFifo: out STD_LOGIC;

    writeHighToFifo: out STD_LOGIC;
    writeLowToFifo: out STD_LOGIC;
    writeToCtrl: out STD_LOGIC;
	 writeToTiming: out STD_LOGIC;
	 
--	 testpin: out STD_LOGIC;
				
    Reset: in  STD_LOGIC;
    clk : in  STD_LOGIC ); 
  end component;
  
  component dataReg Port (
	 A: in STD_LOGIC_VECTOR(31 downto 0); -- mux=00
	 B: in STD_LOGIC_VECTOR(31 downto 0); -- mux=01
	 C: in STD_LOGIC_VECTOR(31 downto 0); -- mux=10
    load: in STD_LOGIC;
	 mux: in STD_LOGIC_VECTOR(1 downto 0);	 
	 output: out STD_LOGIC_VECTOR(31 downto 0);
	 reset: in  STD_LOGIC;
    clk : in  STD_LOGIC );
  end component;
  
  component fifo Port(
	 cpuIn: in STD_LOGIC_VECTOR(31 downto 0);
	 ideIn: in STD_LOGIC_VECTOR(15 downto 0);

	 cpuRd, cpuWrHi, cpuWrLo: in STD_LOGIC;
	 ideRd, ideWr: in STD_LOGIC;

    hasData: out STD_LOGIC;	 
	 ready: out STD_LOGIC;
	 
	 ideOut: out STD_LOGIC_VECTOR(15 downto 0);
	 cpuOut: out STD_LOGIC_VECTOR(31 downto 0);
	 
	 level: out STD_LOGIC_VECTOR(3 downto 0);

	 reset: in  STD_LOGIC;
    clk : in  STD_LOGIC );
  end component;
  
  component ide Port(
    ctrlIn: in STD_LOGIC_VECTOR(12 downto 0);
    timeIn: in STD_LOGIC_VECTOR(8 downto 0);
    ctrlOut: out STD_LOGIC_VECTOR(12 downto 0);
    timeOut: out STD_LOGIC_VECTOR(8 downto 0);
	 loadDataToCtrl: in STD_LOGIC;
	 loadDataToTiming: in STD_LOGIC;
	 
	 readFromFifo: out STD_LOGIC;
	 writeToFifo: out STD_LOGIC;
	 
	 fifoHasData: in STD_LOGIC;
	 fifoIsReady: in STD_LOGIC;

    -- used/controlled by state-machine
    IDEIORDY: in STD_LOGIC;
	 IDEIOW: out STD_LOGIC;
	 IDEIOR: out STD_LOGIC;

	 IDEIntIn: in STD_LOGIC;
	 IDEIntOut: out STD_LOGIC;

	 ideOE: out STD_LOGIC; 
	 
    -- controlled by driver (via "ctrl") 
    IDEAddrOut : out STD_LOGIC_VECTOR (2 downto 0);
    IDECS0Out  : out STD_LOGIC;
    IDECS1Out  : out STD_LOGIC;
    IDEResetOut: out STD_LOGIC;
	 IDEDMACKOut: out STD_LOGIC;


	 ideTSOut: out STD_LOGIC; 
	
	 reset: in  STD_LOGIC;
    clk : in  STD_LOGIC );
  end component;
  
  signal dataOut: STD_LOGIC_VECTOR(31 downto 0);
  signal fifoCPUOut: STD_LOGIC_VECTOR(31 downto 0);
  signal fifoIDEOut: STD_LOGIC_VECTOR(15 downto 0);
  signal fifoLevel: STD_LOGIC_VECTOR(3 downto 0);
  signal state: STD_LOGIC_VECTOR(31 downto 0);
  
  signal timing: STD_LOGIC_VECTOR(8 downto 0); 
  signal ctrl: STD_LOGIC_VECTOR(12 downto 0); 

  signal cpuOE: STD_LOGIC;
  signal ideOE: STD_LOGIC;
  signal loadDataToReg: STD_LOGIC;
  signal dataMux: STD_LOGIC_VECTOR(1 downto 0);
  
  signal cpuReadFromFifo: STD_LOGIC;
  signal cpuWriteHighToFifo: STD_LOGIC;
  signal cpuWriteLowToFifo: STD_LOGIC;
  signal cpuWriteToCtrl: STD_LOGIC;
  signal cpuWriteToTiming: STD_LOGIC;

  signal ideReadFromFifo: STD_LOGIC;
  signal ideWriteToFifo: STD_LOGIC;
  
  signal fifoHasData: STD_LOGIC;
  signal fifoIsReady: STD_LOGIC;
  signal ideTS: STD_LOGIC;

  signal IDEAddrOut : STD_LOGIC_VECTOR (2 downto 0);
  signal IDECS0Out  : STD_LOGIC;
  signal IDECS1Out  : STD_LOGIC;
  signal IDEResetOut: STD_LOGIC;
  signal IDEDMACKOut: STD_LOGIC;

  
begin
  Reset <= '1'; -- GSR pin is not used
  
  Led_Output <= Led_Input;  
  
  cpuData <= dataOut     when (cpuOE='1') else (others => 'Z');
  ideData <= fifoIDEOut  when ((ideOE='1') and (ideTS='0')) else (others => 'Z');
  
  IDEAddr  <= IDEAddrOut when (ideTS='0') else (others => 'Z');
  IDECS0   <= IDECS0Out when (ideTS='0') else 'Z';
  IDECS1   <= IDECS1Out when (ideTS='0') else 'Z';
  IDEReset <= IDEResetOut;
  IDEDMACK <= IDEDMACKOut when (ideTS='0') else 'Z';
      
  state(31 downto 28) <= fifoLevel;
  state(27) <= IDEIORDY;
  state(26) <= IDEIntRq;
  state(25) <= IDEDMARq;
  state(24 downto 16)  <= timing;
  state(15) <= '0';
  state(14) <= '0';
  state(13) <= '0'; 
  state(12 downto 0)  <= ctrl(12 downto 0);
   
  dataC: dataReg port map (
    A      => cpuData,
	 B      => fifoCPUOut,
	 C      => state,
	 load   => loadDataToReg,
	 mux    => dataMux,
	 output => dataOut,
	 reset  => reset, 
	 clk    => clk );
  
  fifoC: fifo port map ( 
    cpuIn   => dataOut,
	 ideIn   => ideData,	 
	 cpuRd   => cpuReadFromFifo,
	 cpuWrHi => cpuWriteHighToFifo,
	 cpuWrLo => cpuWriteLowToFifo,
	 ideRd   => ideReadFromFifo,
	 ideWr   => ideWriteToFifo,
	 ideOut  => fifoIDEOut,
	 cpuOut  => fifoCPUOut,
	 level   => fifoLevel,
	 hasData => fifoHasData,
	 ready   => fifoIsReady,
    reset   => reset, 
	 clk     => clk );
       
  cpuc: cpustates port map ( 
    MemCmd       => MemCmd,
	 Addr         => CPUAddr,
	 dataOE       => cpuOE,	 
	 dataMux      => dataMux,
	 dataLoad     => loadDataToReg, 
	 readFromFifo => cpuReadFromFifo,
    writeHighToFifo => cpuWriteHighToFifo,
    writeLowToFifo  => cpuWriteLowToFifo,
    writeToCtrl     => cpuWriteToCtrl,
	 writeToTiming   => cpuWriteToTiming,


    reset        => reset, 
	 clk          => clk );
	 
	idec: ide port map (
	 ctrlIn         => dataOut(12 downto 0),
	 timeIn         => dataOut(24 downto 16),
	 ctrlOut        => ctrl,
	 timeOut        => timing,
	 loadDataToCtrl => cpuWriteToCtrl,
	 loadDataToTiming=>cpuWriteToTiming,
    readFromFifo   => ideReadFromFifo,
    writeToFifo    => ideWriteToFifo,	 
	 
	 fifoHasData    => fifoHasData,
	 fifoIsReady    => fifoIsReady,
	 
	 IDEIntIn       => IDEIntrq,
	 IDEIntOut      => CPUInt,
	 
	 IDEIORDY       => IDEIORDY,
	 IDEIOW         => IDEIOW,
	 IDEIOR         => IDEIOR,
	 
	 IDEAddrOut     => IDEAddrOut,
    IDECS0Out      => IDECS0Out,
    IDECS1Out      => IDECS1Out,
    IDEResetOut    => IDEResetOut,
	 IDEDMACKOut    => IDEDMACKOut,
	 ideOE          => ideOE,
	 ideTSOut       => ideTS,
	 

	 
    reset          => reset, 
	 clk            => clk );
	 
end Behavioral;
